from pybayesprism.prism import *
import warnings

def get_fraction(bp, which_theta, state_or_type):
    # assert isinstance(bp, BayesPrism)
    assert isinstance(which_theta, str) and which_theta in ["first", "final"]
    assert isinstance(state_or_type, str) and state_or_type in ["state", "type"]
    if which_theta == "first" and state_or_type == "state":
        return bp.posterior_initial_cellState.theta
    if which_theta == "first" and state_or_type == "type":
        return bp.posterior_initial_cellType.theta
    if which_theta == "final":
        if state_or_type == "state":
            warnings.warn("Warning: only cell type is available for updated Gibbs. Returning cell type info.")
        return bp.posterior_theta_f.theta


def get_exp(bp, state_or_type, cell_name):
    assert isinstance(bp, BayesPrism)
    assert isinstance(state_or_type, str) and state_or_type in ["state", "type"]
    if state_or_type == "state":
        return bp.posterior_initial_cellState.Z.loc[:,:,cell_name]
    if state_or_type == "type":
        return bp.posterior_initial_cellType.Z.loc[:,:,cell_name]



from pybayesprism import prism
import numpy as np
import pandas as pd


def step1(bk, sc, cell_type, cell_state):
    r_bk = pd.read_csv("./compare/step1_r_bk.csv", index_col = 0).astype('int32')
    r_sc = pd.read_csv("./compare/step1_r_sc.csv", index_col = 0).astype('int32')
    r_cell_type = list(pd.read_csv("./compare/step1_r_celltype.csv").iloc[:,0])
    r_cell_state = list(pd.read_csv("./compare/step1_r_cellstate.csv").iloc[:,0])

    # bk
    assert np.array_equal(bk.to_numpy(), r_bk.to_numpy()), "bk values"
    assert list(bk.index) == list(r_bk.index), "bk index"
    assert list(bk.columns) == list(r_bk.columns), "bk columns"

    # sc
    assert np.array_equal(sc.to_numpy(), r_sc.to_numpy()), "sc values"
    assert list(sc.index) == list(r_sc.index), "sc index"
    assert list(sc.columns) == list(r_sc.columns), "sc columns"

    # cell_type
    assert cell_type == r_cell_type, "cell_type"

    # cell_state
    assert cell_state == r_cell_state, "cell_state"


def step2(p):
    r_cellstate = pd.read_csv("./compare/step2_r_cellstate.csv", index_col = 0)
    r_celltype = pd.read_csv("./compare/step2_r_celltype.csv", index_col = 0)
    r_mixture = pd.read_csv("./compare/step2_r_mixture.csv", index_col = 0)

    # p.phi_cellState.phi
    assert np.allclose(p.phi_cellState.phi.to_numpy(), r_cellstate.to_numpy()), "phi_cellState.phi values"
    assert list(p.phi_cellState.phi.index) == list(r_cellstate.index), "phi_cellState.phi index"
    assert list(p.phi_cellState.phi.columns) == list(r_cellstate.columns), "phi_cellState.phi columns"
    assert p.phi_cellState.pseudo_min == 1e-08, "phi_cellState.pseudo_min"

    # p.phi_cellType.phi
    assert np.allclose(p.phi_cellType.phi.to_numpy(), r_celltype.to_numpy()), "phi_cellType.phi values"
    assert list(p.phi_cellType.phi.index) == list(r_celltype.index), "phi_cellType.phi index"
    assert list(p.phi_cellType.phi.columns) == list(r_celltype.columns), "phi_cellType.phi columns"
    assert p.phi_cellType.pseudo_min == 1e-08, "phi_cellType.pseudo_min"

    # p.map
    assert sum([len(v) for v in p.map.values()]) == 73, "map"

    # p.key
    assert p.key == "tumor", "key"
    
    # p.mixture
    assert np.allclose(p.mixture.to_numpy(), r_mixture.to_numpy()), "mixture values"
    assert list(p.mixture.index) == list(r_mixture.index), "mixture index"
    assert list(p.mixture.columns) == list(r_mixture.columns), "mixture columns"


def step3(bp):
    r_cs_phi = pd.read_csv("./compare/step3_r_phi.csv", index_col = 0)
    r_cs_Z_index = list(pd.read_csv("./compare/step3_r_cs_Z_index.csv", header=None).iloc[:,0])
    r_cs_Z_columns = list(pd.read_csv("./compare/step3_r_cs_Z_columns.csv", header=None).iloc[:,0])
    r_cs_Z_layer = list(pd.read_csv("./compare/step3_r_cs_Z_layer.csv", header=None).iloc[:,0])
    r_cs_theta = pd.read_csv("./compare/step3_r_cs_theta.csv", index_col = 0)
    r_cs_theta_cv = pd.read_csv("./compare/step3_r_cs_theta_cv.csv", index_col = 0)

    r_ct_Z_melt = pd.read_csv("./compare/step3_r_ct_Z.csv", index_col = 0)
    r_ct_Z = np.empty((np.max(r_ct_Z_melt.iloc[:, 0]), np.max(r_ct_Z_melt.iloc[:, 1]), np.max(r_ct_Z_melt.iloc[:, 2])))
    r_ct_Z_indices = (r_ct_Z_melt.iloc[:, 0:3].to_numpy() - 1).astype(int)
    r_ct_Z_indices_values = r_ct_Z_melt.iloc[:, 3].to_numpy()
    r_ct_Z[r_ct_Z_indices[:, 0], r_ct_Z_indices[:, 1], r_ct_Z_indices[:, 2]] = r_ct_Z_indices_values

    r_ct_Z_index = list(pd.read_csv("./compare/step3_r_ct_Z_index.csv", header=None).iloc[:,0])
    r_ct_Z_columns = list(pd.read_csv("./compare/step3_r_ct_Z_columns.csv", header=None).iloc[:,0])
    r_ct_Z_layer = list(pd.read_csv("./compare/step3_r_ct_Z_layer.csv", header=None).iloc[:,0])
    r_ct_theta = pd.read_csv("./compare/step3_r_ct_theta.csv", index_col = 0)

    # bp.posterior_initial_cellState.Z
    assert np.corrcoef(bp.prism.phi_cellState.phi.to_numpy().flatten(), r_cs_phi.to_numpy().flatten())[0, 1] > 0.999, "cs phi values"
    assert list(bp.posterior_initial_cellState.Z.coords['bulk_id'].values) == list(r_cs_Z_index), "cs Z index"
    assert list(bp.posterior_initial_cellState.Z.coords['gene_id'].values) == list(r_cs_Z_columns), "cs Z columns"
    assert list(bp.posterior_initial_cellState.Z.coords['cell_type'].values) == list(r_cs_Z_layer), "cs Z layer"

    # bp.posterior_initial_cellState.theta
    assert np.corrcoef(bp.posterior_initial_cellState.theta.to_numpy().flatten(), r_cs_theta.to_numpy().flatten())[0, 1] > 0.999, "cs theta values"
    assert list(bp.posterior_initial_cellState.theta.index) == list(r_cs_theta.index), "cs theta index"
    assert list(bp.posterior_initial_cellState.theta.columns) == list(r_cs_theta.columns), "cs theta columns"

    # bp.posterior_initial_cellState.theta_cv
    assert np.corrcoef(bp.posterior_initial_cellState.theta_cv.to_numpy().flatten(), r_cs_theta_cv.to_numpy().flatten())[0, 1] > 0.900, "cs theta_cv values"
    assert list(bp.posterior_initial_cellState.theta_cv.index) == list(r_cs_theta_cv.index), "cs theta_cv index"
    assert list(bp.posterior_initial_cellState.theta_cv.columns) == list(r_cs_theta_cv.columns), "cs theta_cv columns"

    # bp.posterior_initial_cellState.constant
    assert bp.posterior_initial_cellState.constant == 0, "cs constant"

    # bp.posterior_initial_cellType.Z
    assert np.corrcoef(bp.posterior_initial_cellType.Z.to_numpy().flatten(), r_ct_Z.flatten())[0, 1] > 0.999, "ct Z values"
    assert list(bp.posterior_initial_cellType.Z.coords['bulk_id'].values) == list(r_ct_Z_index), "ct Z index"
    assert list(bp.posterior_initial_cellType.Z.coords['gene_id'].values) == list(r_ct_Z_columns), "ct Z columns"
    assert list(bp.posterior_initial_cellType.Z.coords['cell_type_merged'].values) == list(r_ct_Z_layer), "ct Z layer"

    # bp.posterior_initial_cellType.theta
    assert np.corrcoef(bp.posterior_initial_cellType.theta.to_numpy().flatten(), r_ct_theta.to_numpy().flatten())[0, 1] > 0.999, "ct theta values"
    assert list(bp.posterior_initial_cellType.theta.index) == list(r_ct_theta.index), "ct theta index"
    assert list(bp.posterior_initial_cellType.theta.columns) == list(r_ct_theta.columns), "ct theta columns"


def step4(bp):
    r_reference_psi_mal = pd.read_csv("./compare/step4_r_reference_psi_mal.csv", index_col = 0)
    r_reference_psi_env = pd.read_csv("./compare/step4_r_reference_psi_env.csv", index_col = 0)
    r_posterior_theta = pd.read_csv("./compare/step4_r_posterior_theta.csv", index_col = 0)
    r_posterior_theta_cv = pd.read_csv("./compare/step4_r_posterior_theta_cv.csv", index_col = 0)

    # bp.reference_update.psi_mal
    assert np.corrcoef(bp.reference_update.psi_mal.to_numpy().flatten(), r_reference_psi_mal.to_numpy().flatten())[0, 1] > 0.999, "reference psi_mal values"
    assert list(bp.reference_update.psi_mal.index) == list(r_reference_psi_mal.index), "reference psi_mal index"
    assert list(bp.reference_update.psi_mal.columns) == list(r_reference_psi_mal.columns), "reference psi_mal columns"

    # bp.reference_update.psi_env
    assert np.corrcoef(bp.reference_update.psi_env.to_numpy().flatten(), r_reference_psi_env.to_numpy().flatten())[0, 1] > 0.999, "reference psi_env values"
    assert list(bp.reference_update.psi_env.index) == list(r_reference_psi_env.index), "reference psi_env index"
    assert list(bp.reference_update.psi_env.columns) == list(r_reference_psi_env.columns), "reference psi_env columns"

    # bp.reference_update.key
    assert bp.reference_update.key == "tumor", "reference key"

    # bp.posterior_theta_f.theta
    assert np.corrcoef(bp.posterior_theta_f.theta.to_numpy().flatten(), r_posterior_theta.to_numpy().flatten())[0, 1] > 0.999, "posterior theta values"
    assert list(bp.posterior_theta_f.theta.index) == list(r_posterior_theta.index), "posterior theta index"
    assert list(bp.posterior_theta_f.theta.columns) == list(r_posterior_theta.columns), "posterior theta columns"

    # bp.posterior_theta_f.theta_cv
    assert np.corrcoef(bp.posterior_theta_f.theta_cv.to_numpy().flatten(), r_posterior_theta_cv.to_numpy().flatten())[0, 1] > 0.900, "posterior theta_cv values"
    assert list(bp.posterior_theta_f.theta_cv.index) == list(r_posterior_theta_cv.index), "posterior theta_cv index"
    assert list(bp.posterior_theta_f.theta_cv.columns) == list(r_posterior_theta_cv.columns), "posterior theta_cv columns"



import numpy as np
import multiprocessing
import xarray as xr 
from scipy.optimize import minimize
import pandas as pd

from pybayesprism import process_input
from pybayesprism import references as rf


####################################################



def logsumexp(x):
    y = np.max(x)
    return y + np.log(np.sum(np.exp(x - y)))


def log_posterior_gamma(gamma_t, args):
    phi_t = args["phi_t"]
    phi_t_log = args["phi_t_log"]
    Z_gt_t = args["Z_gt_t"]
    Z_t_t = args["Z_t_t"]
    prior_num = args["prior_num"]

    x = phi_t_log + gamma_t
    psi_t_log = x - logsumexp(x)
    log_likelihood = np.nansum(Z_gt_t * psi_t_log)
    log_prior = np.sum(prior_num * gamma_t ** 2)
    log_posterior = log_likelihood + log_prior
    return -1 * log_posterior

def log_posterior_gamma_grad(gamma_t, args):
    phi_t = args["phi_t"]
    phi_t_log = args["phi_t_log"]
    Z_gt_t = args["Z_gt_t"]
    Z_t_t = args["Z_t_t"]
    prior_num = args["prior_num"]

    psi_t = transform_phi_t(phi_t, gamma_t)
    log_likelihood_grad = Z_gt_t - (Z_t_t * psi_t)
    log_prior_grad = 2 * prior_num * gamma_t
    log_posterior_grad = log_likelihood_grad + log_prior_grad
    return -1 * log_posterior_grad

def optimize_psi_multi(phi_t, pgi_t_log, Z_gt_t, Z_t_t, prior_num):
    args = {"phi_t" : phi_t,
            "phi_t_log" : pgi_t_log,
            "Z_gt_t" : Z_gt_t, 
            "Z_t_t" : Z_t_t,
            "prior_num": prior_num}
    res = minimize(fun = log_posterior_gamma, x0 = np.zeros(phi_t.shape[0]), 
                    args = args, method = "CG", jac = log_posterior_gamma_grad)
    value =  [log_posterior_gamma(x, args = args) for x in res.x] 
    return (res.x, value)


def optimize_psi(phi, Z_gt, prior_num, opt_control):
    phi_dim = (phi.index, phi.columns)
    phi = phi.to_numpy()
    phi_log = np.log(phi)
    Z_gt = Z_gt.to_numpy()
    Z_t = np.sum(Z_gt, axis = 0)

    print("running with " + str(opt_control['n.cores']) + " cores!")
    star_input = [(phi[i,:], phi_log[i,:], Z_gt[:,i], Z_t[i], prior_num) for i in range(phi.shape[0])]
    with multiprocessing.Pool(processes = opt_control['n.cores']) as pool:
        results = pool.starmap(optimize_psi_multi, star_input)

    opt_gamma = np.vstack([res for (res, val) in results])
    value = np.sum([val for (res, val) in results])
    
    opt_gamma[np.apply_along_axis(np.max, 1, np.abs(opt_gamma)) > 20, :] = 0
    
    phi = pd.DataFrame(phi, index = phi_dim[0], columns = phi_dim[1])
    psi = transform_phi(phi, opt_gamma)
    psi = pd.DataFrame(psi.to_numpy(), index = phi_dim[0], columns = phi_dim[1])

    return {"psi": psi, "value": value}



####################################################



def transform_phi_transpose(phi_transpose, gamma):
    psi = np.empty((phi_transpose.shape[1], phi_transpose.shape[0]))
    for t in range(psi.shape[0]):
        psi[t, :] = transform_phi_t(phi_transpose[:, t], gamma)
    psi = xr.DataArray(psi, 
                       corrds = [phi_transpose.coords[phi_transpose.dims[1]], 
                                 phi_transpose.coords[phi_transpose.dims[0]]],
                       dims=[phi_transpose.dims[1], phi_transpose.dims[0]])
    return psi

def log_mle_gamma(gamma, args):
    phi_transpose, phi_log_transpose, Z_tg, Z_t = args
    x = phi_log_transpose + gamma
    psi_log = x.T - np.apply_along_axis(logsumexp, 1, x)
    log_likelihood = np.sum(Z_tg * psi_log)
    return -log_likelihood

def log_mle_gamma_grad(gamma, args):
    phi_transpose, phi_log_transpose, Z_tg, Z_t = args
    psi = transform_phi_transpose(phi_transpose, gamma)
    log_likelihood_grad = np.sum(Z_tg - (Z_t * psi), axis = 0)
    return -log_likelihood_grad



def optimize_psi_oneGamma(phi, Z_gt, opt_control, optimizer = "Rcgmin"):
    opt_control["ncores"] = None
    
    if optimizer == "Rcgmin":
        method = "CG"
    elif optimizer == "BFGS":
        method = "BFGS"

    args = (phi.T, np.log(phi).T, Z_gt.T, np.sum(Z_gt, axis = 0))

    opt_res = minimize(fun = log_mle_gamma, 
                       x0 = np.zeros(phi.shape[1]), 
                       args = args, 
                       method = method, 
                       jac = log_mle_gamma_grad)
    
    opt_gamma = opt_res.x
    value = [log_mle_gamma(x, args) for x in opt_gamma]

    psi = transform_phi(phi, np.vstack([opt_gamma] * phi.shape[0]))
    psi = xr.DataArray(psi, 
                       corrds = [phi.coords[phi.dims[1]], 
                                 phi.coords[phi.dims[0]]],
                       dims=[phi.dims[1], phi.dims[0]])

    return {"psi": psi, "value": value, "gamma": opt_gamma}



####################################################



def transform_phi_t(phi_t, gamma_t):
    stabilizing_constant = np.max(gamma_t)
    gamma_stab = gamma_t - stabilizing_constant
    psi_t = phi_t * np.exp(gamma_stab)
    psi_t = psi_t / np.sum(psi_t)
    return psi_t


def transform_phi(phi, gamma):
    psi = pd.DataFrame(np.zeros_like(phi), index = phi.index, columns = phi.columns)
    for t in range(psi.shape[0]):
        psi.iloc[t, :] = transform_phi_t(phi.iloc[t, :], gamma[t, :])
    return psi


def get_MLE_psi_mal(Z_ng_mal, pseudo_min):
    row_sums = Z_ng_mal.sum(axis = 1)
    mle_psi_mal = Z_ng_mal.div(row_sums, axis = 0)
    mle_psi_mal_pseudo_min = process_input.norm_to_one(mle_psi_mal, pseudo_min)
    return mle_psi_mal_pseudo_min


def xr_to_pd(xr):
    assert xr.ndim == 2
    return pd.DataFrame(xr.to_numpy(), 
            index = list(xr.coords[xr.dims[0]].values),
            columns = list(xr.coords[xr.dims[1]].values))


def update_reference(Z, phi_prime, map, key, opt_control, optimizer=["MAP", "MLE"]):
    print("Update the reference matrix ...")
    
    sigma = opt_control["sigma"]
    opt_control.pop("sigma")
    
    optimizer = opt_control["optimizer"]
    opt_control.pop("optimizer")
    
    if key is None:
        Z_gt = np.sum(Z, axis = 0)
        Z_gt = xr_to_pd(Z_gt)

        if optimizer == "MAP":
            psi = optimize_psi(phi = phi_prime.phi, 
                               Z_gt = Z_gt, 
                               prior_num = -1 / (2 * sigma ** 2), 
                               opt_control = opt_control)["psi"]
        if optimizer == "MLE":
            psi = optimize_psi_oneGamma(phi = phi_prime.phi, 
                                        Z_gt = Z_gt, 
                                        opt_control = opt_control)["psi"]
        return rf.RefPhi(psi, None)
    else:
        
        Z_ng_mal = Z.loc[:, :, key]

        Z_ng_mal = xr_to_pd(Z_ng_mal)

        psi_mal = get_MLE_psi_mal(Z_ng_mal, phi_prime.pseudo_min)

        cellType_env = [cell_type for cell_type in map.keys() if cell_type != key]
        Z_gt_env = xr_to_pd(np.sum(Z.loc[:, :, cellType_env], axis = 0))
        phi_env = phi_prime.phi.loc[cellType_env, :]
        
        if optimizer == "MAP":
            psi_env = optimize_psi(phi = phi_env, 
                                   Z_gt = Z_gt_env, 
                                   prior_num = -1 / (2 * sigma ** 2), 
                                   opt_control = opt_control)["psi"]

        if optimizer == "MLE":
            psi_env = optimize_psi_oneGamma(phi = phi_env, 
                                            Z_gt = Z_gt_env, 
                                            opt_control = opt_control)["psi"]
        return rf.RefTumor(psi_mal, psi_env, key, None)





import numpy as np
import scipy
import time
from datetime import datetime, timedelta
import multiprocessing
from itertools import repeat
import tqdm 
import pandas as pd

from pybayesprism import references
from pybayesprism import joint_post
from pybayesprism import theta_post


#https://stackoverflow.com/questions/55818845/fast-vectorized-multinomial-in-python
def multinomial_rvs(count, p):
    """
    Sample from the multinomial distribution with multiple p vectors.

    * count must be an (n-1)-dimensional numpy array.
    * p must an n-dimensional numpy array, n >= 1.  The last axis of p
      holds the sequence of probabilities for a multinomial distribution.

    The return value has the same shape as p.
    """
    out = np.zeros(p.shape, dtype=int)
    ps = p.cumsum(axis=-1)
    # Conditional probabilities
    with np.errstate(divide='ignore', invalid='ignore'):
        condp = p / ps
    condp[np.isnan(condp)] = 0.0
    for i in range(p.shape[-1]-1, 0, -1):
        binsample = np.random.binomial(count, condp[..., i])
        out[..., i] = binsample
        count -= binsample
    out[..., 0] = count
    return out


class GibbsSampler:
    def __init__(self, reference, X, gibbs_control):
        self.reference = reference
        self.X = X
        self.gibbs_control = gibbs_control


    def get_gibbs_idx(gibbs_control):
        chain_length = gibbs_control['chain.length']
        burn_in = gibbs_control['burn.in']
        thinning = gibbs_control['thinning']
        all_idx = np.arange(0, chain_length)
        burned_idx = all_idx[int(burn_in):]
        thinned_idx = burned_idx[np.arange(0, len(burned_idx), thinning)]
        return thinned_idx


    def rdirichlet(alpha):
        x = np.random.gamma(alpha, size = len(alpha))
        return x / np.sum(x)


    def sample_Z_theta_n(X_n, phi, alpha, gibbs_idx, seed = None, compute_elbo = False):
        
        if seed is not None:
            np.random.seed(seed)

        phi = phi.to_numpy()
        G = phi.shape[1]
        K = phi.shape[0]
        
        theta_n_i = np.repeat(1 / K, K)
        Z_n_i = np.empty((G, K))

        Z_n_sum = np.zeros((G, K))
        theta_n_sum = np.zeros(K)
        theta_n2_sum = np.zeros(K)
        
        multinom_coef = 0
        
        # prob_mat = phi * theta_n_i[:, np.newaxis]

        for i in range(np.max(gibbs_idx)):
            prob_mat = phi * theta_n_i[:, np.newaxis]

            # pvals = prob_mat / np.sum(prob_mat, axis=0)
            # Z_n_i = multinomial_rvs(X_n, pvals.T)

            for g in range(G):
                pvals = prob_mat[:, g] / np.sum(prob_mat[:, g])
                Z_n_i[g, :] = np.random.multinomial(n = X_n[g], pvals = pvals)
            
            Z_nk_i = np.sum(Z_n_i, axis = 0)
            theta_n_i = GibbsSampler.rdirichlet(alpha = Z_nk_i + alpha)
            
            if i in gibbs_idx:
                Z_n_sum += Z_n_i
                theta_n_sum += theta_n_i
                theta_n2_sum += theta_n_i**2
                if compute_elbo:
                    multinom_coef += np.sum(np.log(scipy.special.factorial(Z_nk_i))) \
                        - np.sum(np.log(scipy.special.factorial(Z_n_i)))
        
        samples_size = len(gibbs_idx)
        Z_n = Z_n_sum / samples_size
        theta_n = theta_n_sum / samples_size
        theta_cv_n = np.sqrt(theta_n2_sum / samples_size - (theta_n ** 2)) / theta_n
        gibbs_constant = multinom_coef / samples_size
        
        return {'Z_n': Z_n, 
                'theta_n': theta_n, 
                'theta.cv_n': theta_cv_n, 
                'gibbs.constant': gibbs_constant}

    def sample_theta_n(X_n, phi, alpha, gibbs_idx, seed = None):

        if seed is not None:
            np.random.seed(seed)

        phi = phi.to_numpy()
        G = phi.shape[1]
        K = phi.shape[0]

        theta_n_i = np.repeat(1/K, K)
        Z_n_i = np.empty((G, K))

        theta_n_sum = np.zeros(K)
        theta_n2_sum = np.zeros(K)
        
        for i in range(np.max(gibbs_idx)):
            prob_mat = phi * theta_n_i[:, np.newaxis]
            for g in range(G):
                pvals = prob_mat[:, g] / np.sum(prob_mat[:, g])
                Z_n_i[g, :] = np.random.multinomial(n = X_n[g], pvals = pvals)
                
            theta_n_i = GibbsSampler.rdirichlet(alpha = np.sum(Z_n_i, axis = 0) + alpha)
            
            if i in gibbs_idx:
                theta_n_sum += theta_n_i
                theta_n2_sum += theta_n_i**2
        
        samples_size = len(gibbs_idx)
        theta_n = theta_n_sum / samples_size
        theta_cv_n = np.sqrt(theta_n2_sum / samples_size - (theta_n**2)) / theta_n
        
        return {'theta_n': theta_n, 'theta.cv_n': theta_cv_n}


    def my_seconds_to_period(x):
        days = round(x // (60 * 60 * 24))
        hours = round((x - days * 60 * 60 * 24) // (60 * 60))
        minutes = round((x - days * 60 * 60 * 24 - hours * 60 * 60) // 60) + 1
        days_str = '' if days == 0 else str(days) + 'days '
        hours_str = '' if (hours == 0 and days == 0) else str(hours) + 'hrs '
        minutes_str = '' if (minutes == 0 and days == 0 and hours == 0) else str(minutes) + 'mins'
        final_str = days_str + hours_str + minutes_str
        return final_str


    def estimate_gibbs_time(self, final, chain_length = 50):
        ref = self.reference
        X = self.X.to_numpy()
        gibbs_control = self.gibbs_control
        ptm = time.process_time()
        
        if not final:
            assert isinstance(ref, references.RefPhi), "Gibbs is not final but ref is not refPhi"
            GibbsSampler.sample_Z_theta_n(
                X_n = X[0, :], 
                phi = ref.phi,
                alpha = gibbs_control['alpha'],
                gibbs_idx = GibbsSampler.get_gibbs_idx(
                    {'chain.length' : chain_length, 
                     'burn.in' : chain_length * gibbs_control['burn.in'] / gibbs_control['chain.length'], 
                     'thinning' : gibbs_control['thinning']}), 
                seed = gibbs_control['seed'],
                compute_elbo = False)
        else:
            if isinstance(ref, references.RefPhi):
                GibbsSampler.sample_theta_n(
                    X_n = X[0, :], 
                    phi = ref.phi, 
                    alpha = gibbs_control['alpha'], 
                    gibbs_idx = GibbsSampler.get_gibbs_idx(
                        {'chain.length' : chain_length, 
                         'burn.in' : chain_length * gibbs_control['burn.in'] / gibbs_control['chain.length'], 
                         'thinning' : gibbs_control['thinning']}),
                    seed = gibbs_control['seed'])
            if isinstance(ref, references.RefTumor):
                phi_1 = pd.concat([pd.DataFrame(ref.psi_mal.iloc[0, :]).T, ref.psi_env])
                nonzero_idx = np.max(phi_1, axis = 0) > 0
                GibbsSampler.sample_theta_n(
                    X_n = X[0, nonzero_idx], 
                    phi = phi_1.loc[:, nonzero_idx], 
                    alpha = gibbs_control['alpha'], 
                    gibbs_idx = GibbsSampler.get_gibbs_idx(
                        {'chain.length' : chain_length, 
                         'burn.in' : chain_length*gibbs_control['burn.in'] / gibbs_control['chain.length'], 
                         'thinning' : gibbs_control['thinning']}),
                    seed = gibbs_control['seed'])
        
        total_time = time.process_time() - ptm
        estimated_time = gibbs_control['chain.length'] / chain_length * total_time * np.ceil(X.shape[0] / gibbs_control['n.cores']) * 2
        current_time = datetime.now()
        print("Current time: ", current_time)
        print("Estimated time to complete: ", GibbsSampler.my_seconds_to_period(estimated_time))
        print("Estimated finishing time: ", current_time + timedelta(seconds = estimated_time))


    def run_gibbs_refPhi(self, final, compute_elbo):

        assert isinstance(self.reference, references.RefPhi)
        phi = self.reference.phi
        X = self.X.to_numpy()
        gibbs_control = self.gibbs_control
        alpha = gibbs_control['alpha']
        gibbs_idx = GibbsSampler.get_gibbs_idx(gibbs_control)
        seed = gibbs_control['seed']
        print("Start run...")
        
        if not final:
            with multiprocessing.Pool(processes = gibbs_control['n.cores']) as pool:
                X_input = [X[i, :] for i in np.arange(X.shape[0])]
                star_input = zip(X_input, repeat(phi), repeat(alpha), repeat(gibbs_idx), repeat(seed), repeat(compute_elbo))
                gibbs_list = pool.starmap(GibbsSampler.sample_Z_theta_n, tqdm.tqdm(star_input, total=len(X_input)))
            return joint_post.JointPost.new(self.X.index, self.X.columns, phi.index, gibbs_list)
        else:
            with multiprocessing.Pool(processes = gibbs_control['n.cores']) as pool:
                X_input = [X[i, :] for i in np.arange(X.shape[0])]
                star_input = zip(X_input, repeat(phi), repeat(alpha), repeat(gibbs_idx), repeat(seed))
                gibbs_list = pool.starmap(GibbsSampler.sample_theta_n , tqdm.tqdm(star_input, total=len(X_input)))
            return theta_post.ThetaPost.new(self.X.index, self.X.columns, gibbs_list)


    def run_gibbs_refTumor(self):

        assert isinstance(self.reference, references.RefTumor)
        psi_mal = self.reference.psi_mal
        psi_env = self.reference.psi_env
        key = self.reference.key
        X = self.X.to_numpy()
        gibbs_control = self.gibbs_control
        alpha = gibbs_control['alpha']
        gibbs_idx = GibbsSampler.get_gibbs_idx(gibbs_control)
        seed = gibbs_control['seed']
        print("Start run...")
 
        star_input = []
        for i in range(X.shape[0]):
            psi_mal_n = pd.DataFrame(psi_mal.iloc[i, :]).T
            phi_n = pd.concat([psi_mal_n, psi_env])
            nonzero_idx = np.max(phi_n, axis = 0) > 0
            star_input.append((X[i, nonzero_idx], phi_n.loc[:, nonzero_idx], alpha, gibbs_idx, seed))

        with multiprocessing.Pool(processes = gibbs_control['n.cores']) as pool:
            gibbs_list = pool.starmap(GibbsSampler.sample_theta_n , star_input)
        
        return theta_post.ThetaPost.new(self.X.index, [key] + list(psi_env.index), gibbs_list)


    def run(self, final, if_estimate = True, compute_elbo = False):
        if final:
            print("Run Gibbs sampling using updated reference ...")
        else:
            print("Run Gibbs sampling...")
        
        if if_estimate:
            self.estimate_gibbs_time(final = final)
        if isinstance(self.reference, references.RefPhi):
            return GibbsSampler.run_gibbs_refPhi(self, final = final, compute_elbo = compute_elbo)
        if isinstance(self.reference, references.RefTumor):
            return GibbsSampler.run_gibbs_refTumor(self)



import numpy as np
import pandas as pd
from pybayesprism.gibbs import GibbsSampler
from pybayesprism.joint_post import JointPost
from pybayesprism import process_input
import pybayesprism.references as rf
from pybayesprism import optim


class Prism:
    def __init__(self, phi_cellState, phi_cellType, map_, key, mixture):
        self.phi_cellState = phi_cellState
        self.phi_cellType = phi_cellType
        self.map = map_
        self.key = key
        self.mixture = mixture
    
    def valid_opt_control(control):
        ctrl = {'maxit': 100000, 
                'maximize': False, 
                'trace': 0, 
                'eps': 1e-07,
                'dowarn': True, 
                'tol': 0, 
                'maxNA': 500, 
                'n.cores': 1, 
                'optimizer': 'MAP', 
                'sigma': 2}
        
        namc = list(control.keys())

        for name in namc:
            if name in ctrl.keys():
                ctrl[name] = control[name]
            else:
                raise ValueError("Unknown names in opt.control: {}".format(name))

        if ctrl['optimizer'] not in ["MAP", "MLE"]:
            raise ValueError("unknown names of optimizer: " + ctrl['optimizer'])
        
        if ctrl['optimizer'] == "MAP":
            if not isinstance(ctrl['sigma'], (int, float)):
                raise ValueError("sigma needs to be a numeric variable")
            else:
                if ctrl['sigma'] < 0:
                    raise ValueError("sigma needs to be positive")
        return ctrl


    def valid_gibbs_control(control):
        ctrl = {'chain.length': 1000, 
                'burn.in': 500, 
                'thinning': 2, 
                'n.cores': 1, 
                'seed': 123, 
                'alpha': 1}

        namc = list(control.keys())
       
        for name in namc:
            if name in ctrl.keys():
                ctrl[name] = control[name]
            else:
                raise ValueError("Unknown names in opt.control: {}".format(name))

        if ctrl['alpha'] < 0:
            raise ValueError("alpha needs to be positive")
        
        return ctrl


    def new(reference, input_type, cell_type_labels, cell_state_labels, \
            key, mixture, outlier_cut=0.01, outlier_fraction=0.1, pseudo_min=1E-8):
        
        if cell_state_labels is None:
            cell_state_labels = cell_type_labels

        print("number of cells in each cell state")
        print(pd.Series(cell_state_labels).value_counts().sort_values(ascending = False))
        if np.min(pd.Series(cell_state_labels).value_counts()) < 20:
            print("recommend to have sufficient number of cells in each cell state")
        
        if key is None:
            print("No tumor reference is speficied. Reference cell types are treated equally.")
        
        if len(cell_type_labels) != len(cell_state_labels):
            raise ValueError("Error: length of cell.type.labels and cell.state.labels do not match!")
        if len(cell_type_labels) != reference.shape[0]:
            raise ValueError("Error: length of cell.type.labels and nrow(reference) do not match!")
        
        type_to_state_mat = pd.DataFrame({"cell.type.labels": cell_type_labels, "cell.state.labels": cell_state_labels})
        type_to_state_mat = type_to_state_mat.drop_duplicates()
        if type_to_state_mat["cell.state.labels"].value_counts().max() > 1:
            raise ValueError("Error: one or more cell states belong to multiple cell types!")
        if len(pd.unique(cell_type_labels)) > len(pd.unique(cell_state_labels)):
            raise ValueError("Error: more cell types than states!")
        
        if not isinstance(mixture, pd.DataFrame):
            mixture = pd.DataFrame(mixture)
        if not isinstance(reference, pd.DataFrame):
            reference = pd.DataFrame(reference)

        if mixture.shape[1] == 1:
            mixture = mixture.T
            mixture = mixture.rename(index = {0: ["mixture-1"]})

        if mixture.index.equals(pd.RangeIndex(mixture.shape[0])):
            mixture.index = [f"mixture-{i}" for i in range(len(mixture))]
        
        process_input.validate_input(reference)
        process_input.validate_input(mixture)
        
        mixture = process_input.filter_bulk_outlier(mixture, outlier_cut, outlier_fraction)
        
        reference = reference.loc[:, np.sum(reference, axis = 0) > 0]
        
        _, gene_index, _ = np.intersect1d(reference.columns, mixture.columns, return_indices = True)
        gene_index.sort()
        gene_shared = [list(reference.columns)[i] for i in gene_index]
        if len(gene_shared) == 0:
            raise ValueError("Error: gene names of reference and mixture do not match!")
        if len(gene_shared) < 100:
            print("Warning: very few gene from reference and mixture match! Please double check your gene names.")
        
        ref_cs = process_input.collapse(reference, cell_state_labels)
        ref_ct = process_input.collapse(reference, cell_type_labels)
        
        print("Aligning reference and mixture...")
        ref_ct = ref_ct.loc[:, gene_shared]
        ref_cs = ref_cs.loc[:, gene_shared]
        mixture = mixture.loc[:, gene_shared]
        
        print("Normalizing reference...")
        ref_cs = process_input.norm_to_one(ref_cs, pseudo_min)
        ref_ct = process_input.norm_to_one(ref_ct, pseudo_min)
        
        map_ = {cell_type: list(set(cell_state_labels[i] for i, ct in enumerate(cell_type_labels) if ct == cell_type)) for cell_type in ref_ct.index}

        return Prism(
            rf.RefPhi(ref_cs, pseudo_min),
            rf.RefPhi(ref_ct, pseudo_min),
            map_,
            key,
            mixture
        )

    def run(self, n_cores = 1, update_gibbs = True, gibbs_control = {}, opt_control = {}):
        
        if 'n.cores' not in gibbs_control:
            gibbs_control['n.cores'] = n_cores
        if 'n.cores' not in opt_control:
            opt_control['n.cores'] = n_cores
        
        assert isinstance(update_gibbs, bool)
        assert isinstance(n_cores, int)
        
        opt_control = Prism.valid_opt_control(opt_control)
        gibbs_control = Prism.valid_gibbs_control(gibbs_control)
        
        if self.phi_cellState.pseudo_min == 0:
            gibbs_control['alpha'] = max(1, gibbs_control['alpha'])
        
        gibbsSampler_ini_cs = GibbsSampler(reference = self.phi_cellState, 
                                           X = self.mixture, 
                                           gibbs_control = gibbs_control)

        jointPost_ini_cs = gibbsSampler_ini_cs.run(final = False)
        print("Now Merging...")
        jointPost_ini_ct = jointPost_ini_cs.merge_K(map_ = self.map)

        if not update_gibbs:
            bp = BayesPrism(prism = self, 
                                posterior_initial_cellState = jointPost_ini_cs,
                                posterior_initial_cellType = jointPost_ini_ct,
                                control_param = {'gibbs.control': gibbs_control,
                                                 'opt.control': opt_control,
                                                 'update.gibbs': update_gibbs})
            return bp
        else:
            psi = optim.update_reference(Z = jointPost_ini_ct.Z, 
                                         phi_prime = self.phi_cellType, 
                                         map = self.map,
                                         key = self.key, 
                                         opt_control = opt_control)
            
            gibbsSampler_update = GibbsSampler(reference = psi, 
                                               X = self.mixture, 
                                               gibbs_control = gibbs_control)
            
            theta_f = gibbsSampler_update.run(final = True)

            bp = BayesPrism(prism = self, 
                                posterior_initial_cellState = jointPost_ini_cs,
                                posterior_initial_cellType = jointPost_ini_ct, 
                                control_param={'gibbs.control': gibbs_control,
                                               'opt.control': opt_control,
                                               'update.gibbs': update_gibbs},
                                reference_update = psi,
                                posterior_theta_f = theta_f)
            return bp




class BayesPrism:
    def __init__(self, prism, posterior_initial_cellState, posterior_initial_cellType,
                 control_param, reference_update = None, posterior_theta_f = None):
        self.prism = prism
        self.posterior_initial_cellState = posterior_initial_cellState
        self.posterior_initial_cellType = posterior_initial_cellType
        self.control_param = control_param
        self.reference_update = reference_update
        self.posterior_theta_f = posterior_theta_f


    def update_theta(self, gibbs_control = {}, opt_control = {}):

        gibbs_control_bp = self.control_param['gibbs.control']
        opt_control_bp = self.control_param['opt.control']
        
        if gibbs_control:
            for key, value in gibbs_control.items():
                gibbs_control_bp[key] = value
    
        if opt_control:
            for key, value in opt_control.items():
                opt_control_bp[key] = value

        opt_control = Prism.valid_opt_control(opt_control_bp)
        gibbs_control = Prism.valid_gibbs_control(gibbs_control_bp)
        
        psi = optim.update_reference(Z = self.posterior_initial_cellType.Z, 
                                     phi_prime = self.prism.phi_cellType,
                                     map = self.prism.map, 
                                     key = self.prism.key, 
                                     opt_control = opt_control)
        
        gibbsSampler_update = GibbsSampler(reference = psi, 
                                           X = self.prism.mixture, 
                                           gibbs_control = gibbs_control)
        
        theta_f = gibbsSampler_update.run(final = True)
        
        bp_updated = BayesPrism(prism = self.prism, 
                                posterior_initial_cellState = self.posterior_initial_cellState,
                                posterior_initial_cellType = self.posterior_initial_cellType, 
                                control_param={'gibbs.control': gibbs_control, 
                                               'opt.control': opt_control, 
                                               'update.gibbs': True},
                                reference_update = psi,
                                posterior_theta_f = theta_f)

        return bp_updated


import xarray as xr
import numpy as np
import pandas as pd

class JointPost:
    def __init__(self, Z, theta, theta_cv = None, constant = None): 
        self.Z = Z
        self.theta = theta
        self.theta_cv = theta_cv
        self.constant = constant


    def new(bulk_id, gene_id, cell_type, gibbs_list):
        n = len(bulk_id)
        h = len(gene_id)
        k = len(cell_type)
        assert len(gibbs_list) == n

        Z = np.zeros((n, h, k))
        theta = np.zeros((n, k))
        theta_cv = np.zeros((n, k))

        for i in range(n):
            Z[i,:,:] = gibbs_list[i]['Z_n']
            theta[i,:] = gibbs_list[i]['theta_n']

        if 'theta.cv_n' in gibbs_list[0] and gibbs_list[0]['theta.cv_n'] is not None:
            for i in range(n):
                theta_cv[i,:] = gibbs_list[i]['theta.cv_n']
        
        constant = sum(subdict["gibbs.constant"] for subdict in gibbs_list)

        Z = xr.DataArray(Z, coords=[bulk_id, gene_id, cell_type],
                         dims=['bulk_id', 'gene_id', 'cell_type'])
        theta = pd.DataFrame(theta, index = bulk_id, columns = cell_type)
        theta_cv = pd.DataFrame(theta_cv, index = bulk_id, columns = cell_type)

        return JointPost(Z, theta, theta_cv, constant)


    def merge_K(self, map_ : dict):
        bulk_id = self.Z.coords['bulk_id'].values
        gene_id = self.Z.coords['gene_id'].values
        cell_type = self.Z.coords['cell_type'].values
        cell_type_merged = list(map_.keys())

        n = len(bulk_id)
        g = len(gene_id)
        k = len(cell_type)
        k_merged = len(cell_type_merged)

        assert sum([len(v) for v in map_.values()]) == k

        Z = np.zeros((n, g, k_merged))
        theta = np.zeros((n, k_merged))


        Z = xr.DataArray(Z, coords=[('bulk_id', bulk_id), ('gene_id', gene_id), ('cell_type_merged', cell_type_merged)])
        theta = pd.DataFrame(theta, index = bulk_id, columns = cell_type_merged)

        for i in range(k_merged):
            cell_type_merged_k = cell_type_merged[i]
            cell_types_k = map_[cell_type_merged_k]
            if len(cell_types_k) == 1:
                Z.loc[:, :, cell_type_merged_k] = np.squeeze(self.Z.loc[:, :, cell_types_k])
                theta.loc[:, cell_type_merged_k] = np.squeeze(self.theta.loc[:, cell_types_k])
            else:
                Z.loc[:, :, cell_type_merged_k] = np.sum(self.Z.loc[:, :, cell_types_k], axis = 2)
                theta.loc[:, cell_type_merged_k] = np.sum(self.theta.loc[:, cell_types_k], axis = 1)

        return JointPost(Z, theta)


# In R:
# load("your_dir/tutorial.gbm.rdata") 
# write.csv(bk.dat, file = "./data/bk_dat.csv")
# write.csv(sc.dat, file = "./data/sc_dat.csv")
# write(cell.type.labels, file = "./data/cell_type_labels.csv")
# write(cell.state.labels, file = "./data/cell_state_labels.csv")


import pandas as pd
import numpy as np
import pickle
import gzip
import importlib
import time

from pybayesprism import extract
from pybayesprism import gibbs
from pybayesprism import joint_post
from pybayesprism import optim
from pybayesprism import prism
from pybayesprism import process_input
from pybayesprism import references
from pybayesprism import theta_post
from pybayesprism import compare


def reload():
    # importlib.reload(compare)
    importlib.reload(extract)
    importlib.reload(gibbs)
    importlib.reload(joint_post)
    importlib.reload(optim)
    importlib.reload(prism)
    importlib.reload(process_input)
    # importlib.reload(references)
    importlib.reload(theta_post)



# bk_dat = pd.read_csv("./data/bk_dat.csv", index_col = 0).astype(np.int32)
# sc_dat = pd.read_csv("./data/sc_dat.csv", index_col = 0).astype(np.int32)
# cell_type_labels = list(pd.read_csv("./data/cell_type_labels.csv", header = None).iloc[:,0])
# cell_state_labels = list(pd.read_csv("./data/cell_state_labels.csv", header = None).iloc[:,0])

# with gzip.open("./data/d.pkl.gz", "wb") as f:
#     pickle.dump((bk_dat, sc_dat, cell_type_labels, cell_state_labels), f, protocol = 4) 

def a(comp = False):
    reload()
    global bk_dat, sc_dat_filtered_pc, cell_type_labels, cell_state_labels
    with gzip.open("./data/data.pkl.gz", "rb") as f:
        bk_dat, sc_dat, cell_type_labels, cell_state_labels = pickle.load(f)
    sc_dat_filtered = process_input.cleanup_genes(sc_dat, "count.matrix", "hs", \
                    ["Rb", "Mrp", "other_Rb", "chrM", "MALAT1", "chrX", "chrY"], 5)
    sc_dat_filtered_pc = process_input.select_gene_type(sc_dat_filtered, ["protein_coding"])
    compare.step1(bk_dat, sc_dat_filtered_pc, cell_type_labels, cell_state_labels) if comp else None
    


def b(comp = False):
    reload()
    global my_prism
    my_prism = prism.Prism.new(reference = sc_dat_filtered_pc, 
                               mixture = bk_dat, input_type = "count.matrix", 
                               cell_type_labels = cell_type_labels, 
                               cell_state_labels = cell_state_labels, 
                               key = "tumor", 
                               outlier_cut = 0.01, 
                               outlier_fraction = 0.1)
    compare.step2(my_prism) if comp else None


def c(comp = False):
    reload()
    global bp
    bp = my_prism.run(n_cores = 36, update_gibbs = False)
    compare.step3(bp) if comp else None


def d(comp = False):
    reload()
    global bp_res
    bp_res = bp.update_theta(opt_control = {'n.cores' : 20})
    compare.step4(bp_res) if comp else None




import numpy as np
from itertools import compress
import pandas as pd


def norm_to_one(ref, pseudo_min):
    
    ref_index = ref.index
    ref_columns = ref.columns
    
    ref = ref.to_numpy()
    g = ref.shape[1]

    phi = ref / np.sum(ref, axis = 1, keepdims=True) * (1 - pseudo_min * g) + pseudo_min

    min_value = np.min(ref, axis = 1)
    which_row = min_value > 0
    if np.any(which_row):
            phi[which_row, :] = ref[which_row, :] / np.sum(ref[which_row, :], axis = 1, keepdims=True)
    
    phi = pd.DataFrame(phi, index = ref_index, columns = ref_columns)
    return phi


def collapse(ref, labels):

        assert ref.shape[0] == len(labels), "Error: nrow(ref) and length(labels) do not match!"

        non_na_idx = [x is not None for x in labels]
        if non_na_idx.count(False) > 0:
            print("Warning: NA found in the cell type/state labels. These cells will be excluded!")
        labels = list(compress(labels, non_na_idx))
        ref = ref.loc[non_na_idx, :]

        labels_seen = set()
        labels_uniq = [x for x in labels if not (x in labels_seen or labels_seen.add(x))]

        ref_collapsed = pd.DataFrame()
        for label_i in labels_uniq:
            indices = [label_i == i for i in labels]
            ref_label = ref.loc[indices].to_numpy()
            ref_collapsed = ref_collapsed.append(pd.Series(np.sum(ref_label, axis = 0)), ignore_index=True)
        ref_collapsed.index = labels_uniq
        ref_collapsed.columns = ref.columns

        return ref_collapsed


def validate_input(input):
        input_ndarray = input.to_numpy()
        if np.max(input_ndarray) <= 1:
            print("Warning: input seems to be normalized.")
        elif np.max(input_ndarray) < 20:
            print("Warning: input seems to be log-transformed. Please double \
                  check your input. Log transformation should be avoided")
        
        if np.any(input_ndarray < 0) or np.any(np.isinf(input_ndarray)):
            raise ValueError("Error: input contains negative, non-finite or \
                             non-numeric values. Please double check your \
                             input is unnormalized and untransformed raw count.")
        
        if input.columns.empty:
            raise ValueError("Error: please specify the colnames of mixture / \
                             reference using gene identifiers!")

        if not isinstance(input, pd.DataFrame):
            raise ValueError("Error: the type of mixture and reference need \
                             to be DataFrame!")
        

def filter_bulk_outlier(mixture, outlier_cut, outlier_fraction):

    mixture_ndarray = mixture.to_numpy()

    mixture_norm = mixture_ndarray / np.sum(mixture_ndarray, axis = 1, keepdims=True)

    outlier_idx = np.sum(mixture_norm > outlier_cut, axis = 0) / mixture_norm.shape[0] > outlier_fraction

    mixture = mixture.loc[:, ~outlier_idx]
    print("Number of outlier genes filtered from mixture =", np.sum(outlier_idx))
    return mixture


def assign_category(input_genes, species):
    assert len(species) == 2 and species in ["hs", "mm"]
    
    if species == "hs":
        gene_list = pd.read_table(os.path.join(os.path.split(__file__)[0], "txt", "genelist.hs.new.txt"), sep="\t", header=None)
    if species == "mm":
        gene_list = pd.read_table(os.path.join(os.path.split(__file__)[0], "txt", "genelist.mm.new.txt"), sep="\t", header=None)

    if sum(1 for gene in input_genes if gene[:3] == "ENS") > len(input_genes) * 0.8:
        print("EMSEMBLE IDs detected.")
        input_genes_short = [gene.split(".")[0] for gene in input_genes]
        gene_df = gene_list.iloc[:, [0, 1]]
    else:
        print("Gene symbols detected. Recommend to use EMSEMBLE IDs for more unique mapping.")
        input_genes_short = input_genes
        gene_df = gene_list.iloc[:, [0, 2]]
    
    logic = [np.isin(np.array(input_genes_short), gene_df.loc[gene_df.iloc[:,0]==i, 1]) for i in gene_df.iloc[:, 0].unique()]
    logic = np.column_stack(logic)
    gene_group_matrix = pd.DataFrame(logic)
    gene_group_matrix.columns = gene_df.iloc[:, 0].unique()
    gene_group_matrix.index = input_genes
    return gene_group_matrix


def cleanup_genes(input : pd.DataFrame, input_type, species, gene_group, exp_cells=1):
    
    assert species in ["hs", "mm"]
    a = ["other_Rb", "chrM", "chrX", "chrY", "Rb", "Mrp", "act", "hb", "MALAT1"]
    assert all([g in a for g in gene_group])
    assert input_type in ["GEP", "count.matrix"]
    
    if input_type == "GEP":
        exp_cells = min(exp_cells, 1)
        print("As the input is a collpased GEP, exp.cells is set to min(exp.cells,1)")
    
    category_matrix = assign_category(input_genes = input.columns, species = species)
    category_matrix = category_matrix.loc[:, gene_group]

    print("number of genes filtered in each category: ")
    print(np.sum(category_matrix, axis = 0))

    exclude_idx = np.sum(category_matrix, axis=1) > 0
    print("A total of", exclude_idx.sum(), "genes from", gene_group, "have been excluded")
    input_filtered = input.loc[:, ~exclude_idx]
    
    if exp_cells > 0:
        exclude_lowexp_idx = (input_filtered > 0).sum(axis=0) >= exp_cells
        print("A total of", (~exclude_lowexp_idx).sum(), "gene expressed in fewer than", exp_cells, "cells have been excluded")
        input_filtered = input_filtered.loc[:, exclude_lowexp_idx]
    else:
        print("A total of 0 lowly expressed genes have been excluded")
    
    return input_filtered


def select_gene_type(input : pd.DataFrame, gene_type):
    assert all([g in ["protein_coding", "pseudogene", "lincRNA"] for g in gene_type])
    
    input_genes = input.columns
    gene_tab_path = os.path.join(os.path.split(__file__)[0], "txt", "gencode.v22.broad.category.txt")
    gene_list = pd.read_table(gene_tab_path, sep="\t", header=None)
    
    if sum(1 for gene in input_genes if gene[:3] == "ENS") > len(input_genes) * 0.8:
        print("EMSEMBLE IDs detected.")
        input_genes = [gene.split(".")[0] for gene in input_genes]
        gene_list_7 = gene_list.iloc[:,7].tolist()
        gene_match = [gene_list_7.index(x) for x in input_genes if x in gene_list_7]
        gene_df = gene_list.iloc[gene_match, [7, 8]]
    else:
        print("Gene symbols detected. Recommend to use EMSEMBLE IDs for more unique mapping.")
        gene_list_4 = gene_list.iloc[:,4].tolist()
        gene_match = [gene_list_4.index(x) for x in input_genes if x in gene_list_4]
        gene_df = gene_list.iloc[gene_match, [4, 8]]

    gene_df.columns = ["gene_name", "category"]

    selected_gene_idx = gene_df["category"].isin(gene_type)
    
    print("number of genes retained in each category: ")
    print(gene_df.loc[selected_gene_idx, "category"].value_counts())
    input_filtered = input.loc[:, selected_gene_idx.tolist()]
    return input_filtered



import argparse
import pandas as pd
import pickle
import numpy as np

from pybayesprism import prism

parser = argparse.ArgumentParser(description = 'Python Implementation of BayesPrism')

parser.add_argument('-n', dest = 'ncores', help = 'CPU cores')

parser.add_argument('-x', dest = 'file_x', help = 'Bulk count matrix')
parser.add_argument('-x.type', dest = 'file_x_type', help = 'Bulk file type')

parser.add_argument('-ref', dest = 'file_ref', help = 'Reference count matrix')
parser.add_argument('-ref.file.type', dest = 'file_ref_type', help = 'Reference file type')
parser.add_argument('-ref.data.type', dest = 'data_type', help = 'Reference data type')

parser.add_argument('-species', dest = 'species', choices = ['hs', 'mm'], help = 'Species')
parser.add_argument('-out', dest = 'out_prefix', help = 'Output prefix')

parser.add_argument('-file.cell.type', dest = 'file_cell_type', help = 'file of cell type')


args = parser.parse_args()

# unpack args
ncores = args.ncores
file_x = args.file_x
file_x_type = args.file_x_type
file_ref = args.file_ref
file_ref_type = args.file_ref_type
data_type = args.data_type.lower()
outfile = args.out_prefix
file_cell_type = args.file_cell_type
species = args.species

print("CPU cores:", ncores)
print("Bulk count matrix:", file_x)
print("Bulk file type:", file_x_type)
print("Reference count matrix:", file_ref)
print("Reference file type:", file_ref_type)
print("Reference data type:", data_type)
print("File of cell type:", file_cell_type)
print("Output prefix:", outfile)
print("Species:", species)


if file_x_type not in ["csv", "tsv"]:
    print("! Error: the bulk matrix file supports csv, tsv, xls.")
    raise ValueError("Stop [err1]")

if file_ref_type not in ["csv", "tsv"]:
    print("! Error: the reference file supports csv, tsv, xls")
    raise ValueError("Stop [err2]")

print("1) -------- Load count matrix")

try:
    ## rowname: gene name
    ## colname: sample name
    if file_x_type == "tsv":
        x_sep = "\t"
    elif file_x_type == "csv":
        x_sep = ","
    df_x = pd.read_csv(file_x, sep=x_sep, index_col=0)

    ## rowname: gene name
    ## colname: sample name
    if file_ref_type == "tsv":
        ref_sep = "\t"
    elif file_ref_type == "csv":
        ref_sep = ","
    df_ref = pd.read_csv(file_ref, sep=ref_sep, index_col=0)

    #Cell ID
    #Cell Type(Tumor/Normal)
    #Cell subtype
    #Cell tumor flag(0,1:tumor)
    df_ct = pd.read_csv(file_cell_type, sep=",")
    assert df_ct.shape[1] == 4
except:
    print("Error: files could not be loaded.")

if not df_ref.index.isin(df_x.index).all():
    x_gene = sum(df_ref.index.isin(df_x.index))
    print("* Warning: your reference matrix don't have consistent", \
            "gene annotation with your bulk matrix.", x_gene, "/", \
            len(df_x.index), " genes can be matched.")


filter_chr = ["RB", "chrM", "chrX", "chrY"] if species in ["hs", "mm"] else []
    
cell_state_labels = df_ct.iloc[:,3]
cell_type_labels = df_ct.iloc[:,2]

if df_ref.shape[1] != df_ct.shape[0] \
    or not all(colname in list(df_ct.iloc[:, 0]) for colname in df_ref.columns) \
    or not all(cell_id in df_ref.columns for cell_id in df_ct.iloc[:, 0]):
    x_cell = sum(colname not in df_ct.iloc[:, 0] for colname in df_ref.columns)
    print("* Warning: your reference matrix doesn't have consistent cell ID \
    with your cell profile.", x_cell, "/", df_ct.shape[0], " cells can be \
    matched in your reference matrix.")


# Check cell IDs consistency
if data_type == "scrna":
    idx = [list(df_ct.iloc[:,0]).index(x) if x in list(df_ct.iloc[:,0]) else None for x in df_ref.columns]
elif data_type == "gep":
    idx = [list(df_ct.iloc[:,2]).index(x) if x in list(df_ct.iloc[:,2]) else None for x in df_ref.index]
idx = np.array([i is None for i in idx])

if np.sum(idx) > 0:
    if data_type == "scrna":
        print("* Warning:", np.sum(idx), "cells in your reference matrix \
            are not in your cell profile, e.g.:", df_ref.columns[idx])
    else:
        print("* Warning:", np.sum(idx), "cells in your reference matrix \
            are not in your cell profile, e.g.:", df_ref.index[idx])


if len(idx) - np.sum(idx) < 50:
    print("! Error: the remaining cell number is less than 50, stop.")
    raise ValueError("Stop [err5]")


df_ct = df_ct.loc[~idx]
print(df_ct.iloc[0:5,])

cell_state_labels = df_ct.iloc[:, 2].tolist()
cell_type_labels = df_ct.iloc[:, 1].tolist()

if data_type == "scrna":
    df_ref = df_ref.loc[:, df_ct.iloc[:, 0].tolist()]
if data_type == "gep":
    df_ref = df_ref.loc[df_ct.iloc[:, 2].tolist(), :]

print("* Warning:", df_ct.shape[0], "Cells in your reference matrix are used to proceed the next step.")

# Check if there are normal cells
if (df_ct.iloc[:, 3] == 0).sum() == 0:
    print("! Error: no normal cells (state==0) in your cell profile, stop.")
    raise ValueError("Stop [err7]")

key = None

if (df_ct.iloc[:, 3] == 1).sum() == 0:
    pass
else:
    k = df_ct[df_ct.iloc[:, 3] == 1].iloc[:, 1].unique()
    if len(k) > 1:
        key = 'tumor'
        df_ct.loc[df_ct.iloc[:, 3] == 1, df_ct.columns[1]] = key
        df_ct.loc[df_ct.iloc[:, 3] == 1, df_ct.columns[2]] = df_ct.loc[df_ct.iloc[:, 3] == 1, df_ct.columns[2]] + '_tumor'
    elif len(k) == 1:
        if k[0] == "tumor":
            key = k[0]
        else:
            key = k[0]
    

cell_type_labels = df_ct.iloc[:, 1]
cell_state_labels = df_ct.iloc[:, 2]

print("2) -------- Start bayesPrism")

df_ref.index = df_ref.index.str.upper()
# df_ref.columns = df_ref.columns.str.upper()

df_x.index = df_x.index.str.upper()
# df_x.columns = df_x.columns.str.upper()

######################################################
# ref = pd.read_csv("./testdata/[d]ref.csv", index_col=0)
# x = pd.read_csv("./testdata/[d]x.csv", index_col=0)
# ctl = list(pd.read_csv("./testdata/[d]ctl.csv").iloc[:,0])
# csl = list(pd.read_csv("./testdata/[d]csl.csv").iloc[:,0])

# assert np.all(df_ref.index == ref.index)
# assert np.all(df_ref.columns == ref.columns)
# assert np.allclose(df_ref.to_numpy(), ref.to_numpy())
# assert np.all(df_x.index == x.index)
# assert np.all(df_x.columns == x.columns)
# assert np.allclose(df_x.to_numpy(), x.to_numpy())
# assert np.all(ctl == cell_type_labels)
# assert np.all(ctl == cell_type_labels)

# import pdb;pdb.set_trace()
######################################################
if data_type == "scrna":
    if len(filter_chr) != 0:
        ref_dat_filtered = process_input.cleanup_genes(input = df_ref.T,
                                         input_type = "count.matrix",
                                         species = species,
                                         gene_group = ["other_Rb", "chrM", "chrX", "chrY", "Rb", "Mrp"],
                                         exp_cells = 5)
    else:
        ref_dat_filtered = df_ref.T

    my_prism = prism.Prism.new(reference = ref_dat_filtered,
                         input_type = 'count.matrix',
                         cell_type_labels = cell_type_labels,
                         cell_state_labels = cell_state_labels,
                         key = key,
                         mixture = df_x.T,
                         outlier_cut = 0.01,
                         outlier_fraction = 0.1)
    bp = my_prism.run(n_cores = int(ncores))

if data_type == "gep":
    if len(filter_chr) != 0:
        ref_gep_filtered = cleanup_genes(input = df_ref.T,
                                         input_type = "GEP",
                                         species = species,
                                         gene_group = ["Rb", "Mrp", "other_Rb", "chrM", "MALAT1", "chrX", "chrY"],
                                         exp_cells = 5)
    else:
        ref_gep_filtered = df_ref.T

    my_prism = prism.Prism.new(reference = ref_dat_filtered,
                         input_type = 'GEP',
                         cell_type_labels = cell_type_labels,
                         cell_state_labels = cell_state_labels,
                         key = key,
                         mixture = df_x.T,
                         outlier_cut = 0.01,
                         outlier_fraction = 0.1)
    bp = my_prism.run(n_cores = int(ncores))


with open("bp.pkl", "wb") as f:
    pickle.dump(bp, f)



import numpy as np
import pandas as pd

class ThetaPost:
    def __init__(self, theta, theta_cv):
        self.theta = theta
        self.theta_cv = theta_cv

    def new(bulk_id, cell_type, gibbs_list):
        n = len(bulk_id)
        k = len(cell_type)
        assert len(gibbs_list) == n

        theta = np.zeros((n, k))
        theta_cv = np.zeros((n, k))

        for i in range(n):
            theta[i] = gibbs_list[i]['theta_n']
            theta_cv[i] = gibbs_list[i]['theta.cv_n']

        theta = pd.DataFrame(theta, index = bulk_id, columns = cell_type)
        theta_cv = pd.DataFrame(theta_cv, index = bulk_id, columns = cell_type)

        return ThetaPost(theta, theta_cv)


class RefPhi:
    def __init__(self, phi, pseudo_min):
        self.phi = phi
        self.pseudo_min = pseudo_min


class RefTumor():
    def __init__(self, psi_mal, psi_env, key, pseudo_min):
        self.psi_mal = psi_mal
        self.psi_env = psi_env
        self.key = key
        self.pseudo_min = pseudo_min